/*
 * FUSD: the Framework for User-Space Devices
 *
 * Linux Kernel Module
 *
 * Jeremy Elson  <jelson@circlemud.org>
 * Copyright (c) 2001, Sensoria Corporation
 *
 * $Id: kfusd.c,v 1.73 2002/02/07 07:29:06 jelson Exp $
 */

/*
 * Note on debugging messages: Unexpected errors (i.e., indicators of
 * bugs in this kernel module) should always contain '!'.  Expected
 * conditions, even if exceptional (e.g., the device-driver-provider
 * disappears while a file is waiting for a return from a system call)
 * must NOT contain '!'.
 */

#ifndef __KERNEL__
#define __KERNEL__
#endif

#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif

#include <linux/config.h>
#include <linux/stddef.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/poll.h>
#include <linux/version.h>
#include <asm/atomic.h>
#include <asm/uaccess.h>
#include <asm/ioctl.h>


/* Define this if you want to emit debug messages (adds ~8K) */
/* #define CONFIG_FUSD_DEBUG */

/* Default debug level for FUSD messages.  Has no effect unless
 * CONFIG_FUSD_DEBUG is defined. */
#ifndef CONFIG_FUSD_DEBUGLEVEL
#define CONFIG_FUSD_DEBUGLEVEL 2
#endif

/* Define this to do extra (and time-consuming) sanity checking */
/* #define CONFIG_FUSD_VERYSAFE */

/* Define this to check for memory leaks */
/* #define CONFIG_FUSD_MEMDEBUG */

/* Define this to use the faster wake_up_interruptible_sync instead of
 * the normal wake_up_interruptible.  Note: you can't do this unless
 * you're bulding fusd as part of the kernel (not a module); or you've
 * patched kernel/ksyms.s to add __wake_up_sync in addition to
 * __wake_up. */
/* #define CONFIG_FUSD_USE_WAKEUPSYNC */


/**************************************************************************/

#include "fusd.h"
#include "fusd_msg.h"
#include "kfusd.h"

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
# error "FUSD needs to be compiled against Linux Kernel headers >v2.4.0"
#endif

#ifndef CONFIG_DEVFS_FS
# error "FUSD needs DEVFS enabled in the kernel!"
#endif

static devfs_handle_t fusd_handle = 0;
static int last_version = 1;
static int last_transid = 1;

#ifdef MODULE_LICENSE
MODULE_AUTHOR("Jeremy Elson, Sensoria Corp. (c)2001");
MODULE_LICENSE("GPL");
#endif

/***************************Debugging Support*****************************/

#ifdef CONFIG_FUSD_DEBUG

static int fusd_debug_level = CONFIG_FUSD_DEBUGLEVEL;
MODULE_PARM(fusd_debug_level, "i");

#define BUFSIZE 1000 /* kernel's kmalloc pool has a 1012-sized bucket */

static void rdebug_real(char *fmt, ...)
{
  va_list ap;
  int len;
  char *message;

  if ((message = KMALLOC(BUFSIZE, GFP_KERNEL)) == NULL)
    return;

  /* THIS REALLY SHOULD BE VSNPRINTF.. pray that it's big enough */
  va_start(ap, fmt);
  len = vsprintf(message, fmt, ap);
  va_end(ap);

  if (len > BUFSIZE) {
    printk("WARNING: POSSIBLE KERNEL CORRUPTION; MESSAGE TOO LONG\n");
  } else {
    printk("fusd: %.975s\n", message); /* note msgs are truncated at
                      * ~1000 chars to fit inside the 1024 printk
                      * limit imposed by the kernel */
  }

  KFREE(message);
}

#endif /* CONFIG_FUSD_DEBUG */

/******************** Memory Debugging ************************************/

#ifdef CONFIG_FUSD_MEMDEBUG

#define MAX_MEM_DEBUG 10000

DECLARE_MUTEX(fusd_memdebug_sem);

typedef struct {
  void *ptr;
  int line;
  int size;
} mem_debug_t;

mem_debug_t *mem_debug;

static int fusd_mem_init(void)
{
  int i;

  mem_debug = kmalloc(sizeof(mem_debug_t) * MAX_MEM_DEBUG, GFP_KERNEL);

  if (mem_debug == NULL) {
    RDEBUG(2, "argh - memdebug malloc failed!");
    return -ENOMEM;
  }

  /* initialize */
  for (i = 0; i < MAX_MEM_DEBUG; i++)
    mem_debug[i].ptr = NULL;

  RDEBUG(2, "FUSD memory debugger activated");
  return 0;
}

static void fusd_mem_cleanup(void)
{
  int i;
  int count=0;
  for (i = 0; i < MAX_MEM_DEBUG; i++)
    if (mem_debug[i].ptr != NULL) {
      RDEBUG(0, "memdebug: failed to free memory allocated at line %d (%d b)",
	     mem_debug[i].line, mem_debug[i].size);
      count++;
    }
  if (!count)
    RDEBUG(2, "congratulations - memory debugger is happy!");
  kfree(mem_debug);
}

static void fusd_mem_add(void *ptr, int line, int size)
{
  int i;

  if (ptr==NULL)
    return;

  for (i = 0; i < MAX_MEM_DEBUG; i++) {
    if (mem_debug[i].ptr == NULL) {
      mem_debug[i].ptr = ptr;
      mem_debug[i].line = line;
      mem_debug[i].size = size;
      return;
    }
  }
  RDEBUG(1, "WARNING - memdebug out of space!!!!");
}

static void fusd_mem_del(void *ptr)
{
  int i;
  for (i = 0; i < MAX_MEM_DEBUG; i++) {
    if (mem_debug[i].ptr == ptr) {
      mem_debug[i].ptr = NULL;
      return;
    }
  }
  RDEBUG(2, "WARNING - memdebug is confused!!!!");
}


static void *fusd_kmalloc(size_t size, int type, int line)
{
  void *ptr = kmalloc(size, type);
  down(&fusd_memdebug_sem);
  fusd_mem_add(ptr, line, size);
  up(&fusd_memdebug_sem);
  return ptr;
}

static void fusd_kfree(void *ptr)
{
  down(&fusd_memdebug_sem);
  fusd_mem_del(ptr);
  kfree(ptr);
  up(&fusd_memdebug_sem);
}

static void *fusd_vmalloc(size_t size, int line)
{
  void *ptr = vmalloc(size);
  down(&fusd_memdebug_sem);
  fusd_mem_add(ptr, line, size);
  up(&fusd_memdebug_sem);
  return ptr;
}

static void fusd_vfree(void *ptr)
{
  down(&fusd_memdebug_sem);
  fusd_mem_del(ptr);
  vfree(ptr);
  up(&fusd_memdebug_sem);
}

#endif /* CONFIG_FUSD_MEMDEBUG */


/******************* "VerySafe" Debugging Mode ***************************/


#ifdef CONFIG_FUSD_VERYSAFE

/* In 'verysafe' mode, we maintain a list of currently open & valid
 * fusd devices.  later, whenever we get a pointer to a supposedly
 * valid fusd_dev, we check and make sure it's on the list before
 * acting on it.  note that the version number should also be checked
 * to make sure we don't get fooled by reincarnated pointers. */

static fusd_dev_t *fusd_devices[MAX_FUSD_DEVICES];
static int num_fusd_devices = 0;
DECLARE_MUTEX(fusd_devices_sem);

/* add a device to the valid list (see above) */
static int add_fusd_dev(fusd_dev_t *fusd_dev)
{
  int retval;

  down(&fusd_devices_sem);
  if (num_fusd_devices >= MAX_FUSD_DEVICES) {
    retval = -ENOMEM;
    RDEBUG(1, "out of state space while trying to add a new fusd device!");
  } else {
    fusd_devices[num_fusd_devices++] = fusd_dev;
    retval = 0;
  }
  up(&fusd_devices_sem);

  return retval;
}

/* returns the index of a fusd_dev in the valid list, -1 if not found.
 * DO NOT CALL UNLESS fusd_devices_sem is already LOCKED */
static int find_fusd_dev(fusd_dev_t *fusd_dev)
{
  int i;

  for (i = 0; i < num_fusd_devices; i++)
    if (fusd_devices[i] == fusd_dev)
      break;

  if (i >= num_fusd_devices)
    return -1;
  else
    return i;
}


/* delete a fusd_dev off the valid list */
static void del_fusd_dev(fusd_dev_t *fusd_dev)
{
  int i;

  /* critical section: delete the device off the valid list */
  down(&fusd_devices_sem);
  i = find_fusd_dev(fusd_dev);
  if (i < 0) {
    panic("corrupt fusd_dev list!");
  } else {
    /* replace the one we're deleting with whatever is at the end */
    fusd_devices[i] = fusd_devices[--num_fusd_devices];
  }
  up(&fusd_devices_sem);
}


/* is the given pointer a valid fusd_dev?  returns 1 if yes, 0 if no */
static int fusd_dev_valid(fusd_dev_t *fusd_dev, const char *who, int line)
{
  int retval;

  down(&fusd_devices_sem);
  if (find_fusd_dev(fusd_dev) >= 0) {
    retval = 1;
  } else {
    RDEBUG(0, "INVALID FUSD DEV USED IN %s line %d: THIS SHOULD NEVER HAPPEN!",
	   who, line);
    retval = 0;
  }
  up(&fusd_devices_sem);
  return retval;
}


static int fusd_file_valid(fusd_dev_t *fusd_dev, fusd_file_t *fusd_file,
			   const char *who, int line)
{
  if (find_fusd_file(fusd_dev, fusd_file) >= 0) {
    return 1;
  } else {
    RDEBUG(0, "INVALID FUSD DEV USED IN %s line %d: THIS SHOULD NEVER HAPPEN!",
	   who, line);
    return 0;
  }
}

#endif /* CONFIG_FUSD_VERYSAFE */



/*************************************************************************/
/************** STATE MANAGEMENT AND BOOKKEEPING UTILITIES ***************/
/*************************************************************************/

static inline void init_fusd_msg(fusd_msg_t *fusd_msg)
{
  if (fusd_msg == NULL)
    return;

  memset(fusd_msg, 0, sizeof(fusd_msg_t));
  fusd_msg->magic = FUSD_MSG_MAGIC;
  fusd_msg->cmd = FUSD_FOPS_CALL; /* typical, but can be overwritten */
}

/*
 * free a fusd_msg, and NULL out the pointer that points to that fusd_msg.
 */
static inline void free_fusd_msg(fusd_msg_t **fusd_msg)
{
  if (fusd_msg == NULL || *fusd_msg == NULL)
    return;

  if ((*fusd_msg)->data != NULL) {
      VFREE((*fusd_msg)->data);
      (*fusd_msg)->data = NULL;
    }
  KFREE(*fusd_msg);
  *fusd_msg = NULL;
}


/* adjust the size of the 'files' array attached to the device to
 * better match the number of files.  In all cases, size must be at
 * least MIN_ARRAY_SIZE.  Subject to that constraint: if
 * num_files==array_size, the size is doubled; if
 * num_files<array_size/4, the size is halved.  Array is kept as is if
 * the malloc fails.  Returns a pointer to the new file struct or NULL
 * if there isn't one. */
static fusd_file_t **fusd_dev_adjsize(fusd_dev_t *fusd_dev)
{
  fusd_file_t **old_array;
  int old_size;

  old_array = fusd_dev->files;
  old_size = fusd_dev->array_size;

  /* compute the new size of the array */
  if (fusd_dev->array_size > 4*fusd_dev->num_files)
    fusd_dev->array_size /= 2;
  else if (fusd_dev->array_size == fusd_dev->num_files)
    fusd_dev->array_size *= 2;

  /* respect the minimums and maximums (policy) */
  if (fusd_dev->array_size < MIN_FILEARRAY_SIZE)
    fusd_dev->array_size = MIN_FILEARRAY_SIZE;
  if (fusd_dev->array_size > MAX_FILEARRAY_SIZE)
    fusd_dev->array_size = MAX_FILEARRAY_SIZE;

  /* make sure it's sane */
  if (fusd_dev->array_size < fusd_dev->num_files) {
    RDEBUG(0, "fusd_dev_adjsize is royally screwed up!!!!!");
    return fusd_dev->files;
  }

  /* create a new array.  if successful, copy the contents of the old
   * one.  if not, revert back to the old. */
  fusd_dev->files = KMALLOC(fusd_dev->array_size * sizeof(fusd_file_t *),
			    GFP_KERNEL);
  if (fusd_dev->files == NULL) {
    RDEBUG(2, "malloc failed in fusd_dev_adjsize!");
    fusd_dev->files = old_array;
    fusd_dev->array_size = old_size;
  } else {
    if (fusd_dev->name != NULL)
      RDEBUG(10, "/dev/%s now has space for %d files (had %d)", fusd_dev->name,
	     fusd_dev->array_size, old_size);
    memset(fusd_dev->files, 0, fusd_dev->array_size * sizeof(fusd_file_t *));
    memcpy(fusd_dev->files, old_array,
	   fusd_dev->num_files * sizeof(fusd_file_t *));
    KFREE(old_array);
  }

  return fusd_dev->files;
}


/*
 * DEVICE LOCK MUST BE HELD TO CALL THIS FUNCTION
 * 
 * This function frees a device IF there is nothing left that is
 * referencing it.
 *
 * Specifically, we do not free the device if:
 *   - The driver is still active (i.e. device is not a zombie)
 *   - There are still files with the device open
 *
 * If the device is safe to free, it is removed from the valid list
 * (in verysafe mode only) and freed.
 *
 * Returns:  1 if the device was freed
 *           0 if the device still exists (and can be unlocked)
 */
static int maybe_free_fusd_dev(fusd_dev_t *fusd_dev)
{
  fusd_msgC_t *ptr, *next;

  /* DON'T free the device under conditions listed above */
  if (!fusd_dev->zombie || fusd_dev->num_files)
    return 0;

  /* OK - bombs away!  This fusd_dev_t is on its way out the door! */

#ifdef CONFIG_FUSD_VERYSAFE
  /* delete it off the list of valid devices */
  del_fusd_dev(fusd_dev);
#endif

  /* free any outgoing messages that the device might have waiting */
  for (ptr = fusd_dev->msg_head; ptr != NULL; ptr = next) {
    next = ptr->next;
    FREE_FUSD_MSGC(ptr);
  }

  /* free the device's name */
  if (fusd_dev->name != NULL) {
    RDEBUG(8, "freeing state associated with /dev/%s", fusd_dev->name);
    KFREE(fusd_dev->name);
    fusd_dev->name = NULL;
  }

  /* free the array used to store pointers to fusd_file_t's */
  if (fusd_dev->files != NULL) {
    KFREE(fusd_dev->files);
    fusd_dev->files = NULL;
  }

  /* clear the structure and free it! */
  memset(fusd_dev, 0, sizeof(fusd_dev_t));
  KFREE(fusd_dev);

  MOD_DEC_USE_COUNT;
  return 1;
}


/*
 *
 * DO NOT CALL THIS FUNCTION UNLESS THE DEVICE IS ALREADY LOCKED
 *
 * zombify_device: called when the driver disappears.  Indicates that
 * the driver is no longer available to service requests.  If there
 * are no outstanding system calls waiting for the fusd_dev state, the
 * device state itself is freed.
 *
 */
static void zombify_dev(fusd_dev_t *fusd_dev)
{
  int i;

  if (fusd_dev->zombie) {
    RDEBUG(1, "zombify_device called on a zombie!!");
    return;
  }

  fusd_dev->zombie = 1;

  RDEBUG(3, "/dev/%s turning into a zombie (%d open files)", fusd_dev->name,
	 fusd_dev->num_files);

  /* If there are files holding this device open, wake them up. */
  for (i = 0; i < fusd_dev->num_files; i++) {
    wake_up_interruptible(&fusd_dev->files[i]->file_wait);
    wake_up_interruptible(&fusd_dev->files[i]->poll_wait);
  }
}



/* utility function to find the index of a fusd_file in a fusd_dev.
 * returns index if found, -1 if not found.  ASSUMES WE HAVE A VALID
 * fusd_dev.  fusd_file may be NULL if we are searching for an empty
 * slot. */
static int find_fusd_file(fusd_dev_t *fusd_dev, fusd_file_t *fusd_file)
{
  int i, num_files = fusd_dev->num_files;
  fusd_file_t **files = fusd_dev->files;

  for (i = 0; i < num_files; i++)
    if (files[i] == fusd_file)
      return i;

  return -1;
}


/*
 * DEVICE LOCK MUST BE HELD BEFORE THIS IS CALLED
 *
 * Returns 1 if the device was also freed.  0 if only the file was
 * freed.  If the device is freed, then do not try to unlock it!
 * (Callers: Check the return value before unlocking!)
 */
static int free_fusd_file(fusd_dev_t *fusd_dev, fusd_file_t *fusd_file)
{
  int i;

  /* find the index of the file in the device's file-list... */
  if ((i = find_fusd_file(fusd_dev, fusd_file)) < 0)
    panic("corrupted fusd_dev: releasing a file that we think is closed");
    
  /* ...and remove it (by putting the last entry into its place) */
  fusd_dev->files[i] = fusd_dev->files[--(fusd_dev->num_files)];

  /* there might be an incoming message waiting for a restarted system
   * call.  free it. */
  free_fusd_msg(&fusd_file->msg_in);
  
  /* free state associated with this file */
  memset(fusd_file, 0, sizeof(fusd_file_t));
  KFREE(fusd_file);

  /* reduce the size of the file array if necessary */
  if (fusd_dev->array_size > MIN_FILEARRAY_SIZE &&
      fusd_dev->array_size > 4*fusd_dev->num_files)
    fusd_dev_adjsize(fusd_dev);

  /* renumber the array */
  for (i = 0; i < fusd_dev->num_files; i++)
    fusd_dev->files[i]->index = i;

  /* try to free the device */
  return maybe_free_fusd_dev(fusd_dev);
}


/****************************************************************************/
/********************** CLIENT CALLBACK FUNCTIONS ***************************/
/****************************************************************************/


/*
 * fusd_restart_check: Called from the beginning of most system calls
 * to see if we are restarting a system call.
 *
 * In the common case -- that this is NOT a restarted syscall -- we
 * return 0.
 *
 * In the much less common case, we return ERESTARTSYS, and expect the
 * caller to jump right to its fusd_fops_call() call.
 *
 * In the even LESS (hopefully very rare) case when one PID had an
 * interrupted syscall, but a different PID is the next to do a system
 * call on that file descriptor -- well, we lose.  Clear state of that
 * old syscall out and continue as usual.
 */
static __inline__ int fusd_restart_check(fusd_file_t *fusd_file, int subcmd)
{
  if (fusd_file->sys_restarting_pid == 0)
    return 0;

  if (fusd_file->sys_restarting_pid == current->pid &&
      fusd_file->subcmd_outstanding == subcmd) {
    RDEBUG(4, "pid %d restarting system call with transid %ld", current->pid,
	   fusd_file->transid_outstanding);
    return ERESTARTSYS;
  }

  /* uh oh.  we think we are restarting a syscall but PID does match.
     give up. */
  fusd_fops_call_done(fusd_file);
  free_fusd_msg(&fusd_file->msg_in);
  return 0;
}


static int send_to_dev(fusd_dev_t *fusd_dev, fusd_msg_t *fusd_msg, int locked)
{
  fusd_msgC_t *fusd_msgC;

  /* allocate a container for the message */
  if ((fusd_msgC = KMALLOC(sizeof(fusd_msgC_t), GFP_KERNEL)) == NULL)
    return -ENOMEM;

  memset(fusd_msgC, 0, sizeof(fusd_msgC_t));
  memcpy(&fusd_msgC->fusd_msg, fusd_msg, sizeof(fusd_msg_t));

  if (!locked)
    LOCK_FUSD_DEV(fusd_dev);

  /* put the message in the device's outgoing queue.  */
  if (fusd_dev->msg_head == NULL) {
    fusd_dev->msg_head = fusd_dev->msg_tail = fusd_msgC;
  } else {
    fusd_dev->msg_tail->next = fusd_msgC;
    fusd_dev->msg_tail = fusd_msgC;
  }

  if (!locked)
    UNLOCK_FUSD_DEV(fusd_dev);

  /* wake up the driver, which now has a message waiting in its queue */
  WAKE_UP_INTERRUPTIBLE_SYNC(&fusd_dev->dev_wait);

  return 0;

 zombie_dev:
  KFREE(fusd_msgC);
  return -EPIPE;
}



/*
 * fusd_fops_call_send: send a fusd_msg into userspace.
 *
 * NOTE - we are already holding the lock on fusd_file_arg when this
 * function is called, but NOT the lock on the fusd_dev
 */
static int fusd_fops_call_send(fusd_file_t *fusd_file_arg,
			       fusd_msg_t *fusd_msg)
{
  fusd_dev_t *fusd_dev;
  fusd_file_t *fusd_file;

  /* I check this just in case, shouldn't be necessary. */
  GET_FUSD_FILE_AND_DEV(fusd_file_arg, fusd_file, fusd_dev);

  /* make sure message is sane */
  if ((fusd_msg->data == NULL) != (fusd_msg->datalen == 0)) {
    RDEBUG(2, "fusd_fops_call: data pointer and datalen mismatch!");
    return -EINVAL;
  }

  /* fill the rest of the structure */
  fusd_msg->parm.fops_msg.pid = current->pid;
  fusd_msg->parm.fops_msg.uid = current->uid;
  fusd_msg->parm.fops_msg.gid = current->gid;
  fusd_msg->parm.fops_msg.flags = fusd_file->file->f_flags;
  fusd_msg->parm.fops_msg.offset = fusd_file->file->f_pos;
  fusd_msg->parm.fops_msg.device_info = fusd_dev->private_data;
  fusd_msg->parm.fops_msg.private_info = fusd_file->private_data;
  fusd_msg->parm.fops_msg.fusd_file = fusd_file;
  fusd_msg->parm.fops_msg.transid = atomic_inc_and_ret(&last_transid);

  /* set up certain state depending on if we expect a reply */
  switch (fusd_msg->cmd) {

  case FUSD_FOPS_CALL: /* common case */
    fusd_msg->parm.fops_msg.hint = fusd_file->index;
    fusd_file->transid_outstanding = fusd_msg->parm.fops_msg.transid;
    fusd_file->subcmd_outstanding = fusd_msg->subcmd;
    break;

  case FUSD_FOPS_CALL_DROPREPLY:
    /* nothing needed */
    break;

  case FUSD_FOPS_NONBLOCK:
    fusd_msg->parm.fops_msg.hint = fusd_file->index;
    break;

  default:
    RDEBUG(0, "whoa - fusd_fops_call_send got msg with unknown cmd!");
    break;
  }

  /* now add the message to the device's outgoing queue! */
  return send_to_dev(fusd_dev, fusd_msg, 0);


  /* bizarre errors go straight here */
 invalid_dev:
 invalid_file:
 zombie_dev:
  RDEBUG(0, "fusd_fops_call: got invalid device or file!!!!");
  return -EPIPE;
}


/*
 * fusd_fops_call_wait: wait for a driver to reply to a message
 *
 * NOTE - we are already holding the lock on fusd_file_arg when this
 * function is called, but NOT the lock on the fusd_dev
 */
static int fusd_fops_call_wait(fusd_file_t *fusd_file_arg,
			       fusd_msg_t **fusd_msg_reply)
{
  fusd_dev_t *fusd_dev;
  fusd_file_t *fusd_file;
  fusd_msg_t *fusd_msg;
  int retval;

  /* I check this just in case, shouldn't be necessary. */
  GET_FUSD_FILE_AND_DEV(fusd_file_arg, fusd_file, fusd_dev);

  /* initialize first to tell callers there is no reply (yet) */
  if (fusd_msg_reply != NULL)
    *fusd_msg_reply = NULL;

  /* XXX - TODO - FIXME - race here that should be fixed similar to
   * pipe_wait in fs/pipe.c; acquire the dev lock, check for msg_in,
   * then atomically drop lock and go to sleep */
  if (fusd_file->msg_in == NULL) {
    RDEBUG(10, "pid %d blocking on transid %ld", current->pid,
	   fusd_file->transid_outstanding);
    interruptible_sleep_on(&fusd_file->file_wait);

    /*
     * If we woke up due to a signal -- and not due to a reply message
     * coming in -- then we are in some trouble.  The driver is
     * already processing the request and might have changed some
     * state that is hard to roll back.  So, we'll tell the process to
     * restart the system call, and come back to this point when the
     * system call is restarted.  We need to remember the PID to avoid
     * confusion in case there is another process holding this file
     * descriptor that is also trying to make a call.  */
    if (signal_pending(current)) {
      RDEBUG(5, "blocked pid %d got a signal; sending -ERESTARTSYS",
	     current->pid);
      fusd_file->sys_restarting_pid = current->pid;
      return -ERESTARTSYS;
    }
  }

  /* ok - at this point we are awake due to a message received */

  /* make sure that the device is still valid.  it might have
   * disappeared while we were waiting for the reply. */
  if (fusd_dev->zombie) {
    RDEBUG(2, "fusd_fops_call: /dev/%s disappeared instead of replying",
	   fusd_dev->name);
    goto invalid_reply;
  }

  if (fusd_file->msg_in->cmd != FUSD_FOPS_REPLY ||
      fusd_file->msg_in->subcmd != fusd_file->subcmd_outstanding ||
      fusd_file->msg_in->parm.fops_msg.transid != fusd_file->transid_outstanding ||
      fusd_file->msg_in->parm.fops_msg.fusd_file != fusd_file) {
    RDEBUG(2, "fusd_fops_call: invalid reply!");
    goto invalid_reply;
  }

  fusd_msg = fusd_file->msg_in;
  fusd_file->msg_in = NULL;

  /* copy metadata back from userspace */
  fusd_file->file->f_flags = fusd_msg->parm.fops_msg.flags;
  fusd_dev->private_data   = fusd_msg->parm.fops_msg.device_info;
  fusd_file->private_data  = fusd_msg->parm.fops_msg.private_info;

  /* if everything's okay, return the return value.  if caller is
   * willing to take responsibility for freeing the message itself, we
   * return the message too. */
  retval = fusd_msg->parm.fops_msg.retval;
  if (fusd_msg_reply != NULL) {
    /* NOW TRANSFERRING RESPONSIBILITY FOR FREEING THIS DATA TO THE CALLER */
    *fusd_msg_reply = fusd_msg;
  } else {
    /* free the message ourselves */
    free_fusd_msg(&fusd_msg);
  }

  /* success */
  fusd_fops_call_done(fusd_file);
  return retval;

 invalid_reply:
  free_fusd_msg(&fusd_file->msg_in);
  fusd_fops_call_done(fusd_file);
  return -EPIPE;

  /* bizarre errors go straight here */
 invalid_dev:
 invalid_file:
 zombie_dev:
  RDEBUG(0, "fusd_fops_call: got invalid device or file!!!!");
  return -EPIPE;
}


/* fusd client system call handlers should call this after they call
 * fops_call, to destroy the message that was returned to them. */
static void fusd_fops_call_done(fusd_file_t *fusd_file)
{
  fusd_file->sys_restarting_pid = 0;
  fusd_file->transid_outstanding = -1;
  fusd_file->subcmd_outstanding = -1;
}


/*
 * open() has been called on a registered device.  Note that we need
 * to lock the entire device (both top and bottom) because it's
 * possible we can make changes to the device during this operation
 * (e.g. modifying its file list).  We release the lock on the top
 * part in order to dispatch the open request and get a reply.  */
static int fusd_client_open(struct inode *inode, struct file *file)
{
  int i, retval;
  int device_was_freed = 0;
  fusd_dev_t *fusd_dev;
  fusd_file_t *fusd_file;
  fusd_msg_t fusd_msg;

  GET_FUSD_DEV(file->private_data, fusd_dev);
  /* sigh, there is a race here; if the fusd_client_open starts at the
   * same time as fusd_release (i.e. driver dying), we'll try to
   * acquire a lock that does not exist.  Not sure how to fix this one
   * easily (we'd have to keep a list of all valid devices, like
   * "verysafe" mode), but it's very rare in the wild. */
  LOCK_FUSD_DEV(fusd_dev);

  RDEBUG(3, "got an open for /dev/%s (owned by pid %d) from pid %d",
	 fusd_dev->name, fusd_dev->pid, current->pid);

  /* this shouldn't happen.  maybe i'm insane, but i check anyway. */
  for (i = 0; i < fusd_dev->num_files; i++)
    if (fusd_dev->files[i]->file == file) {
      RDEBUG(1, "warning: fusd_client_open got open for already-open file!?");
      retval = -EIO;
      goto error;
    }

  /* You can't open your own file!  Return -EDEADLOCK if someone tries to.
   *
   * XXX - TODO - FIXME - This should eventually be more general
   * deadlock detection of arbitrary length cycles */
  if (current->pid == fusd_dev->pid) {
    RDEBUG(3, "pid %d tried to open its own device (/dev/%s)",
	   fusd_dev->pid, fusd_dev->name);
    retval = -EDEADLOCK;
    goto error;
  }

  /* make more space in the file array if we need it */
  if (fusd_dev->num_files == fusd_dev->array_size &&
      fusd_dev->array_size < MAX_FILEARRAY_SIZE)
    fusd_dev_adjsize(fusd_dev);

  /* make sure we have room... adjsize may have failed */
  if (fusd_dev->num_files >= fusd_dev->array_size) {
    retval = -ENOMEM;
    RDEBUG(1, "/dev/%s out of state space for open files!", fusd_dev->name);
    goto error;
  }

  /* create state for this file */
  if ((fusd_file = KMALLOC(sizeof(fusd_file_t), GFP_KERNEL)) == NULL) {
    retval = -ENOMEM;
    RDEBUG(1, "yikes!  kernel can't allocate memory");
    goto error;
  }
  memset(fusd_file, 0, sizeof(fusd_file_t));
  init_waitqueue_head(&fusd_file->file_wait);
  init_waitqueue_head(&fusd_file->poll_wait);
  init_MUTEX(&fusd_file->file_sem);
  fusd_file->last_poll_sent = -1;
  fusd_file->magic = FUSD_FILE_MAGIC;
  fusd_file->fusd_dev = fusd_dev;
  fusd_file->fusd_dev_version = fusd_dev->version;
  fusd_file->file = file;

  /* add this file to the list of files managed by the device */
  fusd_file->index = fusd_dev->num_files++;
  fusd_dev->files[fusd_file->index] = fusd_file;

  /* store the pointer to this file with the kernel */
  file->private_data = fusd_file;

  /* send message to userspace and get retval */
  init_fusd_msg(&fusd_msg);
  fusd_msg.subcmd = FUSD_OPEN;

  /* send message to userspace and get the reply.  Device can't be
   * locked during that operation. */

  UNLOCK_FUSD_DEV(fusd_dev);
  retval = fusd_fops_call_send(fusd_file, &fusd_msg);
  if (retval >= 0)
    retval = fusd_fops_call_wait(fusd_file, NULL);
  RAWLOCK_FUSD_DEV(fusd_dev); /* must use rawlock so we don't jump
				 out if the device zombified while
				 we were waiting for a reply */

  /* if retval is negative, throw away state... the file open failed */
  if (retval < 0 || fusd_dev->zombie) {
    RDEBUG(3, "...open failed for /dev/%s (owned by pid %d) from pid %d",
	   fusd_dev->name, fusd_dev->pid, current->pid);
    device_was_freed = free_fusd_file(fusd_dev, fusd_file);
  }

 error: /* errors jump here, success falls through from above */

  /* release locks if the device hasn't disappeared in the meanwhile */
  if (!device_was_freed) {
    UNLOCK_FUSD_DEV(fusd_dev);
  } else {
    RDEBUG(10, "client_open: device disappeared inside of client_open");
  }
    
  return retval;

 zombie_dev:
 invalid_dev:
  RDEBUG(1, "invalid or zombie device passed to client_open");
  return -EPIPE;
}


/* close() has been called on a registered device.  like
 * fusd_client_open, we must lock the entire device. */
static int fusd_client_release(struct inode *inode, struct file *file)
{
  int retval;
  fusd_file_t *fusd_file;
  fusd_dev_t *fusd_dev;
  fusd_msg_t fusd_msg;

  GET_FUSD_FILE_AND_DEV(file->private_data, fusd_file, fusd_dev);
 zombie_dev:

  LOCK_FUSD_FILE(fusd_file);

  RDEBUG(3, "got a close on /dev/%s (owned by pid %d) from pid %d",
	 fusd_dev->name, fusd_dev->pid, current->pid);

  /* If the device driver still exists, tell it the file was closed */
  if (fusd_dev->zombie) {
    retval = -EPIPE;
  } else {
    init_fusd_msg(&fusd_msg);
    fusd_msg.subcmd = FUSD_CLOSE;
    retval = fusd_fops_call_send(fusd_file, &fusd_msg);
    if (retval >= 0)
      retval = fusd_fops_call_wait(fusd_file, NULL);
  }

  /* delete the file off the device's file-list, and free it.  note
   * that device may be a zombie right now and may be freed when we
   * come back from free_fusd_file.  we only release the lock if the
   * device still exists. */
  RAWLOCK_FUSD_DEV(fusd_dev);
  if (!free_fusd_file(fusd_dev, fusd_file)) {
    UNLOCK_FUSD_DEV(fusd_dev);
  }

  return retval;

 invalid_dev:
 invalid_file:
  RDEBUG(1, "got a close on client file from pid %d, INVALID DEVICE!",
	 current->pid);
  return -EPIPE;
}



static ssize_t fusd_client_read(struct file *file , char *buf,
			 size_t count, loff_t *offset)
{
  fusd_dev_t *fusd_dev;
  fusd_file_t *fusd_file;
  fusd_msg_t fusd_msg, *reply = NULL;
  int retval = -EPIPE;

  GET_FUSD_FILE_AND_DEV(file->private_data, fusd_file, fusd_dev);
  LOCK_FUSD_FILE(fusd_file);

  RDEBUG(3, "got a read on /dev/%s (owned by pid %d) from pid %d",
	 fusd_dev->name, fusd_dev->pid, current->pid);

  if (fusd_restart_check(fusd_file, FUSD_READ) == ERESTARTSYS)
    goto restarted_syscall;
    
  /* clear the readable bit of our cached poll state */
  fusd_file->cached_poll_state &= ~(FUSD_NOTIFY_INPUT);

  /* make sure we aren't trying to read too big of a buffer */
  if (count > MAX_RW_SIZE)
    count = MAX_RW_SIZE;

  /* send the message */
  init_fusd_msg(&fusd_msg);
  fusd_msg.subcmd = FUSD_READ;
  fusd_msg.parm.fops_msg.length = count;

  /* send message to userspace */
  if ((retval = fusd_fops_call_send(fusd_file, &fusd_msg)) < 0)
    goto done;

 restarted_syscall:
  /* and wait for the reply */
  retval = fusd_fops_call_wait(fusd_file, &reply);

  /* return immediately in case of error */
  if (retval < 0 || reply == NULL)
    goto done;

  /* adjust the reval if the retval indicates a larger read than the
   * data that was actually provided */
  if (reply->datalen != retval) {
    RDEBUG(1, "warning: /dev/%s driver (pid %d) claimed it returned %d bytes "
	   "on read but actually returned %d", 
	   fusd_dev->name, fusd_dev->pid, retval, reply->datalen);
    retval = reply->datalen;
  }

  /* adjust if the device driver gave us more data than the user asked for
   *     (bad!  bad!  why is the driver broken??? */
  if (retval > count) {
    RDEBUG(1, "warning: /dev/%s driver (pid %d) returned %d bytes on read but "
	   "the user only asked for %d", 
	   fusd_dev->name, fusd_dev->pid, retval, count);
    retval = count;
  }

  /* copy the offset back from the message */
  *offset = reply->parm.fops_msg.offset;

  /* IFF return value indicates data present, copy it back */
  if (retval > 0) {
    if (copy_to_user(buf, reply->data, retval)) {
      retval = -EFAULT;
      goto done;
    }
  }

  /* put additional processing here if needed later */

 done:
  free_fusd_msg(&reply);
  UNLOCK_FUSD_FILE(fusd_file);
  return retval;

 zombie_dev:
 invalid_file:
 invalid_dev:
  RDEBUG(3, "got a read on client file from pid %d, driver has disappeared",
	 current->pid);
  return -EPIPE;
}


static ssize_t fusd_client_write(struct file *file,
    const char *buffer,
    size_t length,
    loff_t *offset)
{
  fusd_dev_t *fusd_dev;
  fusd_file_t *fusd_file;
  fusd_msg_t fusd_msg;
  fusd_msg_t *reply = NULL;
  int retval = -EPIPE;

  GET_FUSD_FILE_AND_DEV(file->private_data, fusd_file, fusd_dev);
  LOCK_FUSD_FILE(fusd_file);

  RDEBUG(3, "got a write on /dev/%s (owned by pid %d) from pid %d",
	 fusd_dev->name, fusd_dev->pid, current->pid);

  if (fusd_restart_check(fusd_file, FUSD_WRITE) == ERESTARTSYS)
    goto restarted_syscall;

  if (length < 0) {
    RDEBUG(2, "fusd_client_write: got invalid length %d", length);
    retval = -EINVAL;
    goto done;
  }

  /* clear the writable bit of our cached poll state */
  fusd_file->cached_poll_state &= ~(FUSD_NOTIFY_OUTPUT);

  if (length > MAX_RW_SIZE)
    length = MAX_RW_SIZE;

  init_fusd_msg(&fusd_msg);

  /* sigh.. i guess zero length writes should be legal */
  if (length > 0) {
    if ((fusd_msg.data = VMALLOC(length)) == NULL) {
      retval = -ENOMEM;
      goto done;
    }

    if (copy_from_user(fusd_msg.data, buffer, length)) {
      retval = -EFAULT;
      goto done;
    }
    fusd_msg.datalen = length;
  }
  
  fusd_msg.subcmd = FUSD_WRITE;
  fusd_msg.parm.fops_msg.length = length;
  
  if ((retval = fusd_fops_call_send(fusd_file, &fusd_msg)) < 0)
    goto done;
 restarted_syscall:
  retval = fusd_fops_call_wait(fusd_file, &reply);

  if (reply != NULL)
    *offset = reply->parm.fops_msg.offset;

  /* all done! */
 done:
  free_fusd_msg(&reply);
  UNLOCK_FUSD_FILE(fusd_file);
  return retval;

 zombie_dev:
 invalid_file:
 invalid_dev:
  RDEBUG(3, "got a read on client file from pid %d, driver has disappeared",
	 current->pid);
  return -EPIPE;
}


static ssize_t fusd_client_ioctl(struct inode *inode, struct file *file,
				 unsigned int cmd, unsigned long arg)
{
  fusd_dev_t *fusd_dev;
  fusd_file_t *fusd_file;
  fusd_msg_t fusd_msg, *reply = NULL;
  int retval = -EPIPE, dir, length;

  GET_FUSD_FILE_AND_DEV(file->private_data, fusd_file, fusd_dev);
  LOCK_FUSD_FILE(fusd_file);

  RDEBUG(3, "got an ioctl on /dev/%s (owned by pid %d) from pid %d",
	 fusd_dev->name, fusd_dev->pid, current->pid);

  dir = _IOC_DIR(cmd);
  length = _IOC_SIZE(cmd);

  if (fusd_restart_check(fusd_file, FUSD_IOCTL) == ERESTARTSYS)
    goto restarted_syscall;

  /* if we're trying to read or write, make sure length is sane */
  if ((dir & (_IOC_WRITE | _IOC_READ)) &&
      (length <= 0 || length > MAX_RW_SIZE))
    {
      RDEBUG(2, "client ioctl got crazy IOC_SIZE of %d", length);
      retval = -EINVAL;
      goto done;
    }

  /* fill the struct */
  init_fusd_msg(&fusd_msg);
  fusd_msg.subcmd = FUSD_IOCTL;
  fusd_msg.parm.fops_msg.cmd = cmd;
  fusd_msg.parm.fops_msg.arg = arg;

  /* get the data if user is trying to write to the driver */
  if (dir & _IOC_WRITE) {
    if ((fusd_msg.data = VMALLOC(length)) == NULL) {
      RDEBUG(2, "can't vmalloc for client ioctl!");
      retval = -ENOMEM;
      goto done;
    }

    if (copy_from_user(fusd_msg.data, (void *) arg, length)) {
      retval = -EFAULT;
      goto done;
    }
    fusd_msg.datalen = length;
  }

  /* send request to the driver */
  if ((retval = fusd_fops_call_send(fusd_file, &fusd_msg)) < 0)
    goto done;
 restarted_syscall:
  /* get the response */
  if ((retval = fusd_fops_call_wait(fusd_file, &reply)) < 0 || reply == NULL)
    goto done;

  /* if user is trying to read from the driver, copy data back */
  if (dir & _IOC_READ) {
    if (reply->data == NULL || reply->datalen != length) {
      RDEBUG(2, "client_ioctl read reply with screwy data (%d, %d)",
	     reply->datalen, length);
      retval = -EIO;
      goto done;
    }
    if (copy_to_user((void *)arg, reply->data, length)) {
      retval = -EFAULT;
      goto done;
    }
  }

  /* all done! */
 done:
  free_fusd_msg(&reply);
  UNLOCK_FUSD_FILE(fusd_file);
  return retval;

 zombie_dev:
 invalid_file:
 invalid_dev:
  RDEBUG(3, "got a read on client file from pid %d, driver has disappeared",
	 current->pid);
  return -EPIPE;
}


/*
 * The design of poll for clients is a bit subtle.
 *
 * We don't want the select() call itself to block, so we keep a cache
 * of the most recently known state supplied by the driver.  The cache
 * is initialized to 0 (meaning: nothing readable/writable).
 *
 * When a poll comes in, we do a non-blocking (!) dispatch of a
 * command telling the driver "This is the state we have cached, reply
 * to this call when the state changes.", and then immediately return
 * the cached state.  We tell the kernel's select to sleep on our
 * poll_wait wait queue.
 *
 * When the driver replies, we update our cached info and wake up the
 * wait queue.  Waking up the wait queue will most likely immediately
 * effect a poll again, in which case we will reply whatever we just
 * cached from the driver.
 * 
 */
static unsigned int fusd_client_poll(struct file *file, poll_table *wait)
{
  fusd_dev_t *fusd_dev;
  fusd_file_t *fusd_file;
  fusd_msg_t fusd_msg;
  int kernel_bits;

  GET_FUSD_FILE_AND_DEV(file->private_data, fusd_file, fusd_dev);

  RDEBUG(3, "got a select on /dev/%s (owned by pid %d) from pid %d, cps=%d",
	 fusd_dev->name, fusd_dev->pid, current->pid,
	 fusd_file->cached_poll_state);

  poll_wait(file, &fusd_file->poll_wait, wait);

  /*
   * If our currently cached poll state is not the same as the
   * most-recently-sent polldiff request, then, dispatch a new
   * request.  (We DO NOT wait for a reply, but just dispatch the
   * request).
   *
   * Also, don't send a new polldiff if the most recent one resulted
   * in an error.
   */
  if (fusd_file->last_poll_sent != fusd_file->cached_poll_state &&
      fusd_file->cached_poll_state >= 0) {
    RDEBUG(2, "sending polldiff request because lps=%d, cps=%d",
	   fusd_file->last_poll_sent, fusd_file->cached_poll_state);
    LOCK_FUSD_FILE(fusd_file);
    init_fusd_msg(&fusd_msg);
    fusd_msg.cmd = FUSD_FOPS_NONBLOCK;
    fusd_msg.subcmd = FUSD_POLL_DIFF;
    fusd_msg.parm.fops_msg.cmd = fusd_file->cached_poll_state;
    if (fusd_fops_call_send(fusd_file, &fusd_msg) == 0) {
      /* poll was dispatched successfully */
      fusd_file->last_poll_sent = fusd_file->cached_poll_state;
    }
    UNLOCK_FUSD_FILE(fusd_file);
  }

  /* now return the state we had cached, converting to bits that have
   * meaning to the kernel */
  kernel_bits = 0;
  if (fusd_file->cached_poll_state > 0) {
    if (fusd_file->cached_poll_state & FUSD_NOTIFY_INPUT)
      kernel_bits |= POLLIN;
    if (fusd_file->cached_poll_state & FUSD_NOTIFY_OUTPUT)
      kernel_bits |= POLLOUT;
    if (fusd_file->cached_poll_state & FUSD_NOTIFY_EXCEPT)
      kernel_bits |= POLLPRI;
  }

  return kernel_bits;

 zombie_dev:
 invalid_dev:
 invalid_file:
  RDEBUG(3, "got a select on client file from pid %d, driver has disappeared",
	 current->pid);
  return POLLPRI;
}



static struct file_operations fusd_client_fops = {
  owner:    THIS_MODULE,
  open:     fusd_client_open,
  release:  fusd_client_release,
  read:     fusd_client_read,
  write:    fusd_client_write,
  ioctl:    fusd_client_ioctl,
  poll:     fusd_client_poll,
};


/*************************************************************************/
/*************************************************************************/
/*************************************************************************/


static fusd_file_t *find_fusd_reply_file(fusd_dev_t *fusd_dev, fusd_msg_t *msg)
{
  /* first, try the hint */
  int i = msg->parm.fops_msg.hint;
  if (i >= 0 &&
      i < fusd_dev->num_files &&
      fusd_dev->files[i] == msg->parm.fops_msg.fusd_file)
    {
      RDEBUG(15, "find_fusd_reply_file: hint worked");
    } else {
      /* hint didn't work, fall back to a search of the whole array */
      i = find_fusd_file(fusd_dev, msg->parm.fops_msg.fusd_file);
      RDEBUG(15, "find_fusd_reply_file: hint failed");
    }

  /* we couldn't find anyone waiting for this message! */
  if (i < 0) {
    return NULL;
  } else {
    return fusd_dev->files[i];
  }
}


/* Process an incoming reply to a message dispatched by
 * fusd_fops_call.  Called by fusd_write when a driver writes to
 * /dev/fusd. */
static int fusd_fops_reply(fusd_dev_t *fusd_dev, fusd_msg_t *msg)
{
  fusd_file_t *fusd_file;

  /* figure out the index of the file we are replying to.  usually
   * very fast (uses a hint) */
  if ((fusd_file = find_fusd_reply_file(fusd_dev, msg)) == NULL) {
    RDEBUG(2, "fusd_fops_reply: got a reply on /dev/%s with no connection",
	   fusd_dev->name);
    goto discard;
  }

  /* make sure this is not an old reply going to an old instance that's gone */
  if (fusd_file->fusd_dev_version != fusd_dev->version ||
      msg->parm.fops_msg.transid != fusd_file->transid_outstanding) {
    RDEBUG(2, "fusd_fops_reply: got an old message, discarding");
    goto discard;
  }

  if (fusd_file->msg_in != NULL) {
    RDEBUG(2, "fusd_fops_reply: incoming message already exists!!");
    goto discard;
  }

  RDEBUG(10, "fusd_fops_reply: /dev/%s completed transid %ld (retval %d)",
	 fusd_dev->name, msg->parm.fops_msg.transid,
	 msg->parm.fops_msg.retval);

  fusd_file->msg_in = msg;
  WAKE_UP_INTERRUPTIBLE_SYNC(&fusd_file->file_wait);
  return 0;

 discard:
  /* special case: if the driver sent back a successful "open", but
   * there is no file that is actually open, we forge a "close" so
   * that the driver can maintain balanced open/close pairs. */
  if (msg->subcmd == FUSD_OPEN && msg->parm.fops_msg.retval == 0) {
    RDEBUG(2, "/dev/%s tried to complete an open for transid %ld, "
	   "forging a close", fusd_dev->name, msg->parm.fops_msg.transid);
    msg->cmd = FUSD_FOPS_CALL_DROPREPLY;
    msg->subcmd = FUSD_CLOSE;
    msg->parm.fops_msg.transid = atomic_inc_and_ret(&last_transid);
    send_to_dev(fusd_dev, msg, 1);
    return 0;
  } else {
    return -EPIPE;
  }
}


/* special function to process responses to POLL_DIFF */
static int fusd_polldiff_reply(fusd_dev_t *fusd_dev, fusd_msg_t *msg)
{
  fusd_file_t *fusd_file;

  /* figure out the index of the file we are replying to.  usually
   * very fast (uses a hint) */
  if ((fusd_file = find_fusd_reply_file(fusd_dev, msg)) == NULL)
    return -EPIPE;

  /* record the poll state returned.  convert all negative retvals to -1. */
  if ((fusd_file->cached_poll_state = msg->parm.fops_msg.retval) < 0)
    fusd_file->cached_poll_state = -1;

  RDEBUG(2, "got updated poll state from /dev/%s driver: %d", fusd_dev->name,
	 fusd_file->cached_poll_state);

  /* since the client has returned the polldiff we sent, set
   * last_poll_sent to -1, so that we'll send a polldiff request on
   * the next select. */
  fusd_file->last_poll_sent = -1;

  /* wake up select's queue so that a new polldiff is generated */
  wake_up_interruptible(&fusd_file->poll_wait);

  return 0;
}


static int fusd_register_device(fusd_dev_t *fusd_dev,
				register_msg_t register_msg)
{
  int error = 0;

  /* make sure args are valid */
  if (fusd_dev == NULL) {
    RDEBUG(0, "fusd_register_device: bug in arguments!");
    return -EINVAL;
  }

  /* user can only register one device per instance */
  if (fusd_dev->handle != 0)
    return -EBUSY;

  /* allocate memory for the name, and copy */
  register_msg.name[FUSD_MAX_NAME_LENGTH] = '\0';
  if ((fusd_dev->name = KMALLOC(strlen(register_msg.name)+1, GFP_KERNEL)) == NULL) {
     RDEBUG(1, "yikes!  kernel can't allocate memory");
     return -ENOMEM;
  }
  strcpy(fusd_dev->name, register_msg.name);

  /* try to register the requested device */
  fusd_dev->handle = devfs_register(NULL, register_msg.name,
				    DEVFS_FL_AUTO_DEVNUM,
				    0, 0,
				    S_IFCHR | register_msg.mode,
				    &fusd_client_fops,
				    fusd_dev);
  /* make sure the registration was successful */
  if (fusd_dev->handle == 0) {
    error = -EIO;
    goto register_failed;
  }

  /* remember the user's private data so we can pass it back later */
  fusd_dev->private_data = register_msg.device_info;

  /* everything ok */
  RDEBUG(2, "pid %d registered /dev/%s v%ld", fusd_dev->pid, fusd_dev->name,
	 fusd_dev->version);
  return 0;

register_failed:
  KFREE(fusd_dev->name);
  fusd_dev->name = NULL;
  return error;
}


/****************************************************************************/
/******************** CONTROL CHANNEL CALLBACK FUNCTIONS ********************/
/****************************************************************************/


/* open() called on /dev/fusd itself */
static int fusd_open(struct inode *inode, struct file *file)
{
  fusd_dev_t *fusd_dev = NULL;
  fusd_file_t **file_array = NULL;

  /* keep the module from being unloaded during initialization! */
  MOD_INC_USE_COUNT;

  /* allocate memory for the device state */
  if ((fusd_dev = KMALLOC(sizeof(fusd_dev_t), GFP_KERNEL)) == NULL)
    goto dev_malloc_failed;
  memset(fusd_dev, 0, sizeof(fusd_dev_t));

  if ((file_array = fusd_dev_adjsize(fusd_dev)) == NULL)
    goto file_malloc_failed;

  init_waitqueue_head(&fusd_dev->dev_wait);
  init_MUTEX(&fusd_dev->dev_sem);
  fusd_dev->magic = FUSD_DEV_MAGIC;
  fusd_dev->version = atomic_inc_and_ret(&last_version);
  fusd_dev->pid = current->pid;
  file->private_data = fusd_dev;

#ifdef CONFIG_FUSD_VERYSAFE
  /* add it to the table of valid devices */
  if (add_fusd_dev(fusd_dev) < 0)
    goto add_dev_failed;
#endif

  RDEBUG(3, "pid %d opened /dev/fusd version %ld", fusd_dev->pid,
	 fusd_dev->version);
  return 0;

#ifdef CONFIG_FUSD_VERYSAFE
 add_dev_failed:
  if (file_array != NULL) {
    KFREE(file_array);
  }
#endif
 file_malloc_failed:
  KFREE(fusd_dev);
 dev_malloc_failed:
  RDEBUG(1, "out of memory in fusd_open!");
  MOD_DEC_USE_COUNT;
  return -ENOMEM;
}


/* close() called on /dev/fusd itself.  destroy the device that
 * was registered by it, if any. */
static int fusd_release(struct inode *inode, struct file *file)
{
  fusd_dev_t *fusd_dev;

  GET_FUSD_DEV(file->private_data, fusd_dev);
  LOCK_FUSD_DEV(fusd_dev);

  if (fusd_dev->pid != current->pid) {
    RDEBUG(2, "yikes!: when releasing device, pid mismatch");
  }

  RDEBUG(3, "pid %d closing /dev/fusd", current->pid);

  if (fusd_dev->handle) {
    if (fusd_dev->name)
      RDEBUG(2, "unregistering /dev/%s from pid %d", fusd_dev->name,
	     fusd_dev->pid);
    devfs_unregister(fusd_dev->handle);
    fusd_dev->handle = 0;
  } else {
    RDEBUG(2, "pid %d releasing without a registered device!", current->pid);
  }

  /* mark the driver as being gone */
  zombify_dev(fusd_dev);

  /* ...and possibly free it.  (Release lock if it hasn't been freed) */
  if (!maybe_free_fusd_dev(fusd_dev))
    UNLOCK_FUSD_DEV(fusd_dev);

  return 0;

 zombie_dev:
 invalid_dev:
  RDEBUG(1, "invalid device found in fusd_release!!");
  return -ENODEV;
}


/*
 * This function processes messages coming from userspace device drivers
 * (i.e., writes to the /dev/fusd control channel.)
 */
static ssize_t fusd_process_write(struct file *file,
   const char *user_msg_buffer, size_t user_msg_len,
   const char *user_data_buffer, size_t user_data_len)
{
  fusd_dev_t *fusd_dev;
  fusd_msg_t *msg = NULL;
  int retval = 0;
  int yield = 0;

  GET_FUSD_DEV(file->private_data, fusd_dev);
  LOCK_FUSD_DEV(fusd_dev);

  /* get the header from userspace (first make sure there's enough data) */
  if (user_msg_len != sizeof(fusd_msg_t)) {
    RDEBUG(6, "control channel got bad write of %d bytes (wanted %d)",
	   user_msg_len, sizeof(fusd_msg_t));
    retval = -EINVAL;
    goto out_no_free;
  }
  if ((msg = KMALLOC(sizeof(fusd_msg_t), GFP_KERNEL)) == NULL) {
    retval = -ENOMEM;
    RDEBUG(1, "yikes!  kernel can't allocate memory");
    goto out;
  }
  memset(msg, 0, sizeof(fusd_msg_t));

  if (copy_from_user(msg, user_msg_buffer, sizeof(fusd_msg_t))) {
    retval = -EFAULT;
    goto out;
  }
  msg->data = NULL; /* pointers from userspace have no meaning */

  /* check the magic number before acting on the message at all */
  if (msg->magic != FUSD_MSG_MAGIC) {
    RDEBUG(2, "got invalid magic number on /dev/fusd write from pid %d",
	   current->pid);
    retval = -EIO;
    goto out;
  }

  /* now get data portion of the message */
  if (user_data_len < 0 || user_data_len > MAX_RW_SIZE) {
    RDEBUG(2, "fusd_process_write: got invalid length %d", user_data_len);
    retval = -EINVAL;
    goto out;
  }
  if (msg->datalen != user_data_len) {
    RDEBUG(2, "msg->datalen(%d) != user_data_len(%d), sigh!",
	   msg->datalen, user_data_len);
    retval = -EINVAL;
    goto out;
  }
  if (user_data_len > 0) {
    if (user_data_buffer == NULL) {
      RDEBUG(2, "msg->datalen and no data buffer, sigh!");
      retval = -EINVAL;
      goto out;
    }
    if ((msg->data = VMALLOC(user_data_len)) == NULL) {
      retval = -ENOMEM;
      RDEBUG(1, "yikes!  kernel can't allocate memory");
      goto out;
    }
    if (copy_from_user(msg->data, user_data_buffer, user_data_len)) {
      retval = -EFAULT;
      goto out;
    }
  }

  /* before device registration, the only command allowed is 'register'. */
  if (!fusd_dev->handle && msg->cmd != FUSD_REGISTER_DEVICE) {
    RDEBUG(2, "got a message other than 'register' on a new device!");
    retval = -EINVAL;
    goto out;
  }

  /* now dispatch the command to the appropriate handler */
  switch (msg->cmd) {
  case FUSD_REGISTER_DEVICE:
    retval = fusd_register_device(fusd_dev, msg->parm.register_msg);
    goto out;
    break;
  case FUSD_FOPS_REPLY:
    /* if reply is successful, DO NOT free the message */
    if ((retval = fusd_fops_reply(fusd_dev, msg)) == 0) {
      yield = 1;
      goto out_no_free;
    }
    break;
  case FUSD_FOPS_NONBLOCK_REPLY:
    switch (msg->subcmd) {
    case FUSD_POLL_DIFF:
      retval = fusd_polldiff_reply(fusd_dev, msg);
      break;
    default:
      RDEBUG(2, "fusd_fops_nonblock got unknown subcmd %d", msg->subcmd);
      retval = -EINVAL;
    }
    break;
  default:
    RDEBUG(3, "warning: unknown message type of %d received!", msg->cmd);
    retval = -EINVAL;
    goto out;
    break;
  }


 out:
  if (msg && msg->data) {
    VFREE(msg->data);
    msg->data = NULL;
  }
  if (msg != NULL) {
    KFREE(msg);
    msg = NULL;
  }

 out_no_free:

  /* the functions we call indicate success by returning 0.  we
   * convert that into a success indication by changing the retval to
   * the length of the write. */
  if (retval == 0)
    retval = user_data_len + user_msg_len;

  UNLOCK_FUSD_DEV(fusd_dev);

  /* if we successfully completed someone's syscall, yield the
   * processor to them immediately as a throughput optimization.  we
   * also hope that in the case of bulk data transfer, their next
   * syscall will come in before we are scheduled again. */
  if (yield) {
    current->policy |= SCHED_YIELD;
    schedule();
  }

  return retval;

 zombie_dev:
 invalid_dev:
  RDEBUG(1, "fusd_process_write: got invalid device!");
  return -EPIPE;
}


static ssize_t fusd_write(struct file *file,
    const char *buffer,
    size_t length,
    loff_t *offset)
{
  return fusd_process_write(file, buffer, length, NULL, 0);
}


static ssize_t fusd_writev(struct file *file,
			   const struct iovec *iov,
			   unsigned long count,
			   loff_t *offset)
{
  if (count != 2) {
    RDEBUG(2, "fusd_writev: got illegal iov count of %ld", count);
    return -EINVAL;
  }

  return fusd_process_write(file,
			    iov[0].iov_base, iov[0].iov_len,
			    iov[1].iov_base, iov[1].iov_len);
}


/* fusd_read: a process is reading on /dev/fusd. return any messages
 * waiting to go from kernel to userspace.
 *
 * Important note: there are 2 possible read modes;
 *   1) header-read mode; just the fusd_msg structure is returned.
 *
 *   2) data-read mode; the data portion of a call (NOT including the
 *   fusd_msg structure) is returned.
 *
 * The protocol this function expects the user-space library to follow
 * is:
 *   1) Userspace library reads header.
 *   2) If fusd_msg->datalen == 0, goto step 4.
 *   3) Userspace library reads data.
 *   4) Message gets dequeued by the kernel.
 *
 * In other words, userspace first reads the header.  Then, if and
 * only if the header you read indicates that data follows, userspace
 * follows with a read for that data.
 *
 * For the header read, the length requested MUST be the exact length
 * sizeof(fusd_msg_t).  The corresponding data read must request
 * exactly the number of bytes in the data portion of the message.  NO
 * OTHER READ LENGTHS ARE ALLOWED - ALL OTHER READ LENGTHS WILL GET AN
 * -EINVAL.  This is done as a basic safety measure to make sure we're
 * talking to a userspace library that understands our protocol, and
 * to detect framing errors.
 *
 * (note: normally you'd have to worry about reentrancy in a function
 * like this because the process can block on the userspace access and
 * another might try to read.  usually we would copy the message into
 * a temp location to make sure two processes don't get the same
 * message.  however in this very specialized case, we're okay,
 * because each instance of /dev/fusd has a completely independent
 * message queue.)  */


/* do a "header" read: used by fusd_read */
static int fusd_read_header(char *user_buffer, size_t user_length, fusd_msg_t *msg)
{
  int len = sizeof(fusd_msg_t);

  if (user_length != len) {
    RDEBUG(4, "bad length of %d sent to /dev/fusd for peek", user_length);
    return -EINVAL;
  }

  if (copy_to_user(user_buffer, msg, len))
    return -EFAULT;

  return sizeof(fusd_msg_t);
}


/* do a "data" read: used by fusd_read */
static int fusd_read_data(char *user_buffer, size_t user_length, fusd_msg_t *msg)
{
  int len = msg->datalen;

  if (len == 0 || msg->data == NULL) {
    RDEBUG(1, "fusd_read_data: no data to send!");
    return -EIO;
  }

  /* make sure the user is requesting exactly the right amount (as a
     sanity check) */
  if (user_length != len) {
    RDEBUG(4, "bad read for %d bytes on /dev/fusd (need %d)",user_length,len);
    return -EINVAL;
  }

  /* now copy to userspace */
  if (copy_to_user(user_buffer, msg->data, len))
	return -EFAULT;

  /* done! */
  return len;
}


static ssize_t fusd_read(struct file *file,
    char *user_buffer,    /* The buffer to fill with data */
    size_t user_length,   /* The length of the buffer */
    loff_t *offset)  /* Our offset in the file */
{
  fusd_dev_t *fusd_dev;
  fusd_msgC_t *msg_out;
  int retval, dequeue = 0;

  GET_FUSD_DEV(file->private_data, fusd_dev);
  LOCK_FUSD_DEV(fusd_dev);

  RDEBUG(15, "driver pid %d (/dev/%s) entering fusd_read", current->pid,
	 fusd_dev->name);

  /* if no messages are waiting, either block or return EAGAIN */
  while ((msg_out = fusd_dev->msg_head) == NULL) {
    if (file->f_flags & O_NONBLOCK) {
      retval = -EAGAIN;
      goto out;
    }

    /* sleep, waiting for a message to arrive.  in most drivers this
     * would be a race condition (what if a packet arrives after the
     * comparison but before we go to sleep?).  in our case there is
     * not a race here (I THINK) because the messages don't arrive on
     * interrupts, they arrive on another system calls (writes from
     * other processes).  */
    UNLOCK_FUSD_DEV(fusd_dev);
    interruptible_sleep_on(&fusd_dev->dev_wait);
    LOCK_FUSD_DEV(fusd_dev);

    /* we're back awake!  --see if a signal woke us up */
    if (signal_pending(current)) {
      retval = -ERESTARTSYS;
      goto out;
    }
  }

  /* is this a header read or data read? */
  if (!msg_out->peeked) {
    /* this is a header read (first read) */
    retval = fusd_read_header(user_buffer, user_length, &msg_out->fusd_msg);

    /* is there data?  if so, make sure next read gets data.  if not,
     * make sure message is dequeued now.*/
    if (msg_out->fusd_msg.datalen) {
      msg_out->peeked = 1;
      dequeue = 0;
    } else {
      dequeue = 1;
    }
  } else {
    /* this is a data read (second read) */
    retval = fusd_read_data(user_buffer, user_length, &msg_out->fusd_msg);
    dequeue = 1; /* message should be dequeued */
  }

  /* if this message is done, take it out of the outgoing queue */
  if (dequeue) {
    if (fusd_dev->msg_tail == fusd_dev->msg_head)
      fusd_dev->msg_tail = fusd_dev->msg_head = NULL;
    else
      fusd_dev->msg_head = msg_out->next;
    FREE_FUSD_MSGC(msg_out);
  }

 out:
  UNLOCK_FUSD_DEV(fusd_dev);
  return retval;

 zombie_dev:
 invalid_dev:
  RDEBUG(2, "got read on /dev/fusd for unknown device!");
  return -EPIPE;
}


/* a poll on /dev/fusd itself (the control channel) */
static unsigned int fusd_poll(struct file *file, poll_table *wait)
{
  fusd_dev_t *fusd_dev;
  GET_FUSD_DEV(file->private_data, fusd_dev);

  poll_wait(file, &fusd_dev->dev_wait, wait);

  if (fusd_dev->msg_head != NULL) {
    return POLLIN | POLLRDNORM;
  }

 invalid_dev:
 zombie_dev:
  return 0;
}

/*************************************************************************/

static struct file_operations fusd_fops = {
  owner:    THIS_MODULE,
  open:     fusd_open,
  read:     fusd_read,
  write:    fusd_write,
  writev:   fusd_writev,
  release:  fusd_release,
  poll:     fusd_poll,
};
  

static int init_fusd(void)
{
#ifdef CONFIG_FUSD_MEMDEBUG
  int retval;

  if ((retval = fusd_mem_init()) < 0)
    return retval;
#endif

  printk(KERN_INFO
	 "fusd: starting, $Revision: 1.73 $, $Date: 2002/02/07 07:29:06 $");
#ifdef CVSTAG
  printk(", release %s", CVSTAG);
#endif
#ifdef CONFIG_FUSD_DEBUG
  printk(", debuglevel=%d\n", fusd_debug_level);
#else
  printk(", debugging messages disabled\n");
#endif

  /* register the control channel with devfs */
  fusd_handle = devfs_register(NULL, "fusd",
			       DEVFS_FL_AUTO_DEVNUM,
			       0, 0,
			       S_IFCHR | 0666,
			       &fusd_fops, NULL);

  if (fusd_handle == NULL) {
    printk("fusd: unable to register fusd control device!\n");
    return -EIO;
  }

  RDEBUG(1, "registration successful");

  return 0;
}

static void cleanup_fusd(void)
{
  devfs_unregister(fusd_handle);
#ifdef CONFIG_FUSD_MEMDEBUG
  fusd_mem_cleanup();
#endif
  RDEBUG(1, "unloading");
}

module_init(init_fusd);
module_exit(cleanup_fusd);
